-- map_video_display
-- @short: Specify virtual object to output mapping.
-- @inargs: vid:src, number:display
-- @inargs: vid:src, number:display, number:blithint
-- @outargs: bool:success
-- @longdescr: This functions updates the mapping between a higher level
-- visual object *src* and an output *display*. This includes the currently
-- set shader, custom texture coordinates and so on.
-- The optional *blithint* argument (default: HINT_NONE) can be one of the
-- following: HINT_FIT, HINT_CROP, HINT_ROTATE_CW_90, HINT_ROTATE_CCW_90,
-- HINT_YFLIP. It is also possible to OR it with the optional HINT_PRIMARY.
-- FIT and CROP determine the strategy to deal with for the case where the
-- current mode on the *display* fail to match the storage dimensions of *src*.
-- YFLIP, ROTATE_CW_90, ROTATE_CCW_90 deals with outputs that can have a
-- varying physical rotation and supports accelerated mapping.
-- HINT_PRIMARY tells the platform layer that the synchronisation to the
-- display SHOULD be prioritized in the case of multiple displays.
-- The function returns *true* if the object was successfully mapped to
-- the display, and *false* if the underlying platform rejected it for some
-- reason, as not all outputs can accept all kinds of sources.
-- @note: A *src* referencing an object with a feed- function, such as
-- one coming from ref:define_recordtarget, ref:define_calctarget and so
-- on, is a terminal state transition.
-- @note: Multiple output paths with varying performance characteristics
-- may apply here, including forced repacking stages that are very costly.
-- Avoid using custom texture coordinate sets, custom shaders or using the
-- *src* object as part of other rendering operations to ensure that any
-- fast scanout path can be applied.
-- @note: A rendertarget as *src* has a different coordinate transform
-- when going from 'as a texture' to 'as an output' behavior and may end
-- up inverted. The engine will attempt to account for this, but only if
-- no custom texture coordinate set has been defined for the object.
-- @group: vidsys
-- @cfunction: videomapping
-- @related:
function main()
#ifdef MAIN
	resize_video_canvas(640, 480);

	img_1 = fill_surface(320, 200, 255, 0, 0);
	img_2 = fill_surface(320, 200, 0, 255, 0);
	img_3 = fill_surface(320, 200, 0, 0, 255);
	img_4 = fill_surface(320, 200, 255, 0, 255);
	show_image({img_1, img_2, img_3, img_4});
	move_image(img_2, 320, 0);
	move_image(img_3, 0, 200);
	move_image(img_4, 320, 200);

	a = null_surface(640, 480);
	move_image(a, 320, 240, 1000);
	image_sharestorage(WORLDID, a);

	outputs = {};
	list = video_displaymodes();

	uniq = 0;
	for i,v in ipairs(list) do
		if (outputs[v.displayid] == nil) then
			outputs[v.displayid] = v;
			uniq = uniq + 1;
		end
	end

	if (uniq < 2) then
		return shutdown("insufficient outputs");
	end

	local first = nil;
	local second = nil;

	for k,v in pairs(list) do
		if (first == nil) then
			first = k;
		elseif (second == nil and k ~= first) then
			second = k;
			break;
		end
	end

	if (second == nil) then
		return shutdown("no secondary display output found.");
	end

	map_video_display(first, WORLDID);
	map_video_display(second, next_a);
#endif

#ifdef ERROR
	map_video_display("random", WORLDID);
#endif
end
